/********************************************************************************
*                                                                               *
*                   E Z   C o l o r   Q u a n t i z a t i o n                   *
*                                                                               *
*********************************************************************************
* Copyright (C) 1999,2024 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* This library is free software; you can redistribute it and/or modify          *
* it under the terms of the GNU Lesser General Public License as published by   *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU Lesser General Public License for more details.                           *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public License      *
* along with this program.  If not, see <http://www.gnu.org/licenses/>          *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxmath.h"
#include "FXElement.h"


/*
  Notes:

  - Use fxezquantize for a quick test to see if the image contains
    less than 256 colors; for example loading then saving back out
    an 8-bit GIF image.  This ensures that the original set of
    colors is maintained.

  - Updated version sorts the colormap colors, smallest values first.
*/


using namespace FX;


/*******************************************************************************/

namespace FX {


extern FXbool fxezquantize(FXuchar* dst,const FXColor* src,FXColor* colormap,FXint& actualcolors,FXint w,FXint h,FXint maxcolors);


#if 0
// EZ quantization may be used if w*h<=maxcolors, or if the actual colors
// used is less than maxcolors; using fxezquantize assures that no
// loss of data occurs repeatedly loading and saving the same file!
FXbool fxezquantize(FXuchar* dst,const FXColor* src,FXColor* colormap,FXint& actualcolors,FXint w,FXint h,FXint maxcolors){
  FXColor  colortable[337];             // Colors encountered in image
  FXushort indextable[337];             // Map index assigned to color
  FXint    npixels=w*h;
  FXint    ncolors=0;
  FXColor  color;
  FXint    i,p,x;

  FXASSERT(maxcolors<=256);

  // Clear map index
  memset(indextable,0xff,sizeof(indextable));

  // Hash all colors from image
  for(i=0; i<npixels; i++){

    // Get pixel
    color=src[i];

    // Find position in table
    p=color%337;
    x=color%331+1;
    while(indextable[p]!=0xffff){       // Empty slot
      if(colortable[p]==color) goto nxt;
      p=(p+x)%337;
      }

    // If no more room in colormap, we failed
    if(ncolors>=maxcolors) return false;

    // Map color to index
    colortable[p]=color;                // Add color to color hash table
    indextable[p]=ncolors;              // Remember map index of this color

    // Add new color
    colormap[ncolors++]=color;          // Add color to color map

    // Next pixel
nxt:continue;
    }

  // Now loop through image, assigning map indices; all colors
  // must be in the map, so each lookup will be successful.
  for(i=0; i<npixels; i++){

    // Get pixel
    color=src[i];

    // Find position in table
    p=color%337;
    x=color%331+1;
    while(colortable[p]!=color){
      p=(p+x)%337;
      }

    // Should be in the list
    FXASSERT(indextable[p]<ncolors);

    // Output map index
    dst[i]=(FXuchar)indextable[p];
    }

  // Actual number of colors used
  actualcolors=ncolors;

  return true;
  }
#endif

// EZ quantization may be used if w*h<=maxcolors, or if the actual colors
// used is less than maxcolors; using fxezquantize assures that no
// loss of data occurs repeatedly loading and saving the same file!
FXbool fxezquantize(FXuchar* dst,const FXColor* src,FXColor* colormap,FXint& actualcolors,FXint w,FXint h,FXint maxcolors){
  FXColor colortable[337];      // Colors encountered in image
  FXuchar indextable[337];      // Map index assigned to color
  FXuchar indexmap[256];        // Back map
  FXint   npixels=w*h;
  FXint   ncolors=0;
  FXColor color;
  FXint   i,p,x;

  FXASSERT(maxcolors<=256);

  // Initialize arrays
  clearElms(colortable,ARRAYNUMBER(colortable));
  clearElms(indextable,ARRAYNUMBER(indextable));

  // Hash all colors from image
  for(i=1; i<npixels; i++){

    // Next pixel's color
    color=src[i];

    // Hash and hash increment
    p=color%337;
    x=color%331+1;

    // Check if color in table
    while(indextable[p]){
      if(colortable[p]==color) goto nxt;
      p=(p+x)%337;
      }

    // If no more room in colormap, we failed
    if(ncolors>=maxcolors) return false;

    // Add to colortable
    colortable[p]=color;
    indextable[p]=1;

    // Find location to insert at
    x=ncolors++;
    while(0<x && color<colormap[x-1]){
      colormap[x]=colormap[x-1];
      indexmap[x]=indexmap[x-1];
      --x;
      }

    // Add to colormap and indexmap
    colormap[x]=color;
    indexmap[x]=p;

    // Next pixel
nxt:continue;
    }

  FXTRACE((500,"ncolors=%d\n",ncolors));

  // Reassign flagged items with colormap index
  for(i=0; i<ncolors; ++i){
    FXASSERT(indextable[indexmap[i]]==1);
    indextable[indexmap[i]]=i;
    }

  // Now loop through image, assigning map indices; all colors
  // must be in the map, so each lookup will be successful.
  for(i=0; i<npixels; i++){

    // Get pixel
    color=src[i];

    // Find position in table
    p=color%337;
    x=color%331+1;
    while(colortable[p]!=color){
      p=(p+x)%337;
      }

    // Should be in the list
    FXASSERT(indextable[p]<ncolors);

    // Output map index
    dst[i]=indextable[p];
    }

  // Actual number of colors used
  actualcolors=ncolors;

  return true;
  }

}
