/********************************************************************************
*                                                                               *
*                         P N G    I n p u t / O u t p u t                      *
*                                                                               *
*********************************************************************************
* Copyright (C) 2023,2024 by Sander Jansen.   All Rights Reserved.              *
*********************************************************************************
* This library is free software; you can redistribute it and/or modify          *
* it under the terms of the GNU Lesser General Public License as published by   *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU Lesser General Public License for more details.                           *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public License      *
* along with this program.  If not, see <http://www.gnu.org/licenses/>          *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxcrc.h"
#include "fxmath.h"
#include "FXElement.h"
#include "FXArray.h"
#include "FXMetaClass.h"
#include "FXHash.h"
#include "FXElement.h"
#include "FXStream.h"
#include "FXString.h"
#ifdef HAVE_ZLIB_H
#include <zlib.h>
#endif


/*
  Notes:

  - Only requirement is zlib library install.

  - Issues RGBGRAY not aware of BE vs LE; r,b, order
    is immaterial, but alpha channel is not....

  - Comparing strings like "tRNS" instead of constant 32-bit
    integers...

  - References:

    http://www.w3.org/TR/REC-png.html
    http://www.graphicswiz.com/png/
    http://www.inforamp.net/~poynton
    http://www.libpng.org/pub/png/
    https://libspng.org/download.html (new!)
*/

#define RGBGRAY(x) ((0xff<<24)|(x<<16)|(x<<8)|(x))

#define CHUNK(c0,c1,c2,c3) ((c3)|((c2)<<8)|((c1)<<16)|((c0)<<24))

#define BUFFERSIZE 256

using namespace FX;

/*******************************************************************************/

namespace FX {


#ifndef FXLOADPNG
extern FXAPI FXbool fxcheckPNG(FXStream& store);
extern FXAPI FXbool fxloadPNG(FXStream& store,FXColor*& data,FXint& width,FXint& height);
extern FXAPI FXbool fxsavePNG(FXStream& store,const FXColor* data,FXint width,FXint height);
extern FXAPI FXbool fxezquantize(FXuchar* dst,const FXColor* src,FXColor* colormap,FXint& actualcolors,FXint w,FXint h,FXint maxcolors);
#endif


#ifdef HAVE_ZLIB_H

// PNG file signature
const FXuchar signature[8]={0x89,0x50,0x4E,0x47,0x0D,0x0A,0x1A,0x0A};

// 1-bit grayscale palette
const FXuchar pal1bit[]={0,255};

// 2-bit grayscale palette
const FXuchar pal2bit[]={0,85,170,255};

// 4-bit grayscale palette
const FXuchar pal4bit[]={0,17,34,51,68,85,102,119,136,153,170,187,204,221,238,255};

// Interleaving x
const FXuint xoffset[] ={0,4,0,2,0,1,0};
const FXuint xstep[]   ={8,8,4,4,2,2,1};
const FXuint xshift[]  ={3,3,2,2,1,1,0};

// Interleaving y
const FXuint yoffset[] ={0,0,4,0,2,0,1};
const FXuint ystep[]   ={8,8,8,4,4,2,2};
const FXuint yshift[]  ={3,3,3,2,2,1,1};

// Filters
const FXuchar Adaptive = 0;

// Compression
const FXuchar Deflate  = 0;

// Interlacing
const FXuchar NoInterlace = 0;
const FXuchar Adam7       = 1;

// Image type
const FXuchar Gray      = 0;
const FXuchar Invalid   = 1;
const FXuchar RGB       = 2;
const FXuchar Indexed   = 3;
const FXuchar GrayAlpha = 4;
const FXuchar RGBA      = 6;

// Chunk codes
const FXuint IHDR = CHUNK('I','H','D','R');
const FXuint tIME = CHUNK('t','I','M','E');
const FXuint zTXt = CHUNK('z','T','X','t');
const FXuint tEXt = CHUNK('t','E','X','t');
const FXuint iTXt = CHUNK('i','T','X','t');
const FXuint pHYs = CHUNK('p','H','Y','s');
const FXuint sPLT = CHUNK('s','P','L','T');
const FXuint eXIf = CHUNK('e','X','I','f');
const FXuint acTL = CHUNK('a','c','T','L');
const FXuint iCCP = CHUNK('i','C','C','P');
const FXuint cICP = CHUNK('c','I','C','P');
const FXuint sRGB = CHUNK('s','R','G','B');
const FXuint sBIT = CHUNK('s','B','I','T');
const FXuint gAMA = CHUNK('g','A','M','A');
const FXuint cHRM = CHUNK('c','H','R','M');
const FXuint mDCv = CHUNK('m','D','C','v');
const FXuint cLLi = CHUNK('c','L','L','i');
const FXuint PLTE = CHUNK('P','L','T','E');
const FXuint tRNS = CHUNK('t','R','N','S');
const FXuint hIST = CHUNK('h','I','S','T');
const FXuint bKGD = CHUNK('b','K','G','D');
const FXuint IDAT = CHUNK('I','D','A','T');
const FXuint fdAT = CHUNK('f','d','A','T');
const FXuint fcTL = CHUNK('f','c','T','L');
const FXuint IEND = CHUNK('I','E','N','D');

/*******************************************************************************/

class PNGBase {
protected:
  FXuint        width;                  // Width
  FXuint        height;                 // Height
  FXuchar       imagetype;              // Image type (Gray, Indexed, ...)
  FXuchar       bitdepth;               // Bit depth per channel
  FXuchar       compression;            // Compression mode
  FXuchar       filter;                 // Filter
  FXuchar       interlace;              // Interlace (NoInterlace, Adam7)
protected:
  FXColor      *colormap;               // Colormap
  FXushort      ncolormap;              // Number of colors in colormap
protected:
  FXuint        nrowbytes;              // Number of bytes in row
  FXuchar       stride;                 // Pixel stride
  FXuchar      *crow;                   // Current row
  FXuchar      *prow;                   // Previous row
protected:
  FXuchar      *buffer;                 // Zlib input/output buffer
  z_stream      stream;                 // Zlib stream
protected:
  FXbool prepare_stride();
  FXbool is_valid() const;
protected:

  // Initialize
  PNGBase():width(0),height(0),imagetype(0),bitdepth(8),compression(Deflate),filter(Adaptive),interlace(NoInterlace),colormap(nullptr),ncolormap(0),nrowbytes(0),stride(0),crow(nullptr),prow(nullptr),buffer(nullptr){
    clearElms(&stream,1);
    }

  // Initialize with size
  PNGBase(FXuint ww,FXuint hh):width(ww),height(hh),imagetype(0),bitdepth(8),compression(Deflate),filter(Adaptive),interlace(NoInterlace),colormap(nullptr),ncolormap(0),nrowbytes(0),stride(0),crow(nullptr),prow(nullptr),buffer(nullptr){
    clearElms(&stream,1);
    }

  // Destroy
 ~PNGBase(){
    freeElms(colormap);
    freeElms(crow);
    freeElms(prow);
    freeElms(buffer);
    }
  };


// Prepare row stride
FXbool PNGBase::prepare_stride(){
  switch(imagetype){
  case Gray:
  case Indexed:
    nrowbytes = (7 + (width * bitdepth)) >> 3;
    stride    = (7 + bitdepth) >> 3;
    return true;
  case GrayAlpha:
    nrowbytes = (width * bitdepth * 2) >> 3;
    stride    = (bitdepth * 2) >> 3;
    return true;
  case RGB:
    nrowbytes = (width * bitdepth * 3) >> 3;
    stride    = (bitdepth * 3) >> 3;
    return true;
  case RGBA:
    nrowbytes = (width * bitdepth * 4) >> 3;
    stride    = (bitdepth * 4) >> 3;
    return true;
    }
  return false;
  }


// Check parameter validity
FXbool PNGBase::is_valid() const {
  if(width==0 || height==0) return false;
  if(compression!=Deflate) return false;
  if(filter!=Adaptive) return false;
  if(interlace > Adam7) return false;
  switch(imagetype){
  case Gray:
    return (bitdepth==1) || (bitdepth==2) || (bitdepth==4) || (bitdepth==8) || (bitdepth==16);
  case Indexed:
    return (bitdepth==1) || (bitdepth==2) || (bitdepth==4) || (bitdepth==8);
  case GrayAlpha:
  case RGB:
  case RGBA:
    return (bitdepth==8) || (bitdepth==16);
    }
  return false;
  }


/*******************************************************************************/

// PNG Decoder
class PNGDecoder : public PNGBase {
protected:
  FXColor      *image;                  // Output Image
protected:
  FXColor       transparency;           // Transparent color for bitdepth < 16
  FXuchar       transparency_16bit[6];  // Transparent color for bitdepth == 16
protected:
  FXuint        irowbytes[7];           // Number of raw bytes in a row for each interlace pass
  FXuint        iheight[7];             // Height of image for each interlace pass
  FXuint        imagerow;               // Current row
  FXuchar       imagepass;              // Current interlace pass
protected:
  void apply_transparency();
  void apply_filter(FXuchar type,FXuchar* row,const FXuchar* prev,const FXint nbytes);
  void copy_pixels(FXColor* dst,const FXColor* end,const FXuchar* src,FXint nbytes,FXint step=1);
protected:
  FXbool process_row();
  FXbool prepare();
  FXbool load_header(FXStream&);
  FXbool load_colormap(FXStream&,FXuint);
  FXbool load_transparency(FXStream&,FXuint);
  FXbool load_data(FXStream&,FXuint);
  FXbool parse(FXStream& store,FXColor*&,FXint&,FXint&);
public:
  PNGDecoder();
  static FXbool load(FXStream& store,FXColor*&,FXint&,FXint&);
 ~PNGDecoder();
  };


PNGDecoder::PNGDecoder():image(nullptr),transparency(0),imagerow(0),imagepass(0){
  }


PNGDecoder::~PNGDecoder(){
  freeElms(image);
  }


FXbool PNGDecoder::load(FXStream& store,FXColor*& img,FXint& ww,FXint& hh){
  PNGDecoder decoder;
  return decoder.parse(store,img,ww,hh);
  }


FXbool PNGDecoder::load_header(FXStream& store){
  FXuchar header[8];
  FXuint storedcrc;
  FXuint chunklength;
  FXuint chunkid;
  FXuint crc=~0;

  // Load signature
  store.load(header,8);
  if(!equalElms(header,signature,ARRAYNUMBER(signature))){
    FXTRACE((100,"fxloadPNG: signature mismatch.\n"));
    return false;
    }

  // Load IHDR chunk
  store >> chunklength;
  store >> chunkid;

  // Check IHDR chunk
  if(chunklength!=13 || chunkid!=IHDR){
    FXTRACE((100,"fxloadPNG: corrupted IHDR.\n"));
    return false;
    }

  store >> width;
  store >> height;
  store >> bitdepth;
  store >> imagetype;
  store >> compression;
  store >> filter;
  store >> interlace;
  store >> storedcrc;

  FXTRACE((100,"png: width %u height %u bitdepth %hhu imagetype %hhu compression %hhu filter %hhu interlace %hhu\n",width,height,bitdepth,imagetype,compression,filter,interlace));

  // Init CRC
  crc=CRC32::CRC(crc,chunkid);
  crc=CRC32::CRC(crc,width);
  crc=CRC32::CRC(crc,height);
  crc=CRC32::CRC(crc,bitdepth);
  crc=CRC32::CRC(crc,imagetype);
  crc=CRC32::CRC(crc,compression);
  crc=CRC32::CRC(crc,filter);
  crc=CRC32::CRC(crc,interlace);
  crc^=~0;

  // Check CRC
  if(storedcrc!=crc){
    FXTRACE((100,"png: crc mismatch %x = %x\n",storedcrc,crc));
    return false;
    }
  return true;
  }


// Load colormap
FXbool PNGDecoder::load_colormap(FXStream& store,FXuint nbytes){
  if((nbytes%3)==0 && nbytes<=3*256){

    // Init the crc
    FXuint savedcrc=0;
    FXuint crc=~0;

    crc=CRC32::CRC(crc,PLTE);

    // Read the colormap
    if(nbytes>0){
      FXuchar r,g,b;
      ncolormap=nbytes/3;
      callocElms(colormap,256);
      for(FXuint c=0; c<ncolormap; c++){
        store >> r;
        store >> g;
        store >> b;
        crc=CRC32::CRC(crc,r);
        crc=CRC32::CRC(crc,g);
        crc=CRC32::CRC(crc,b);
        colormap[c]=FXRGB(r,g,b);
        }
      }
    crc^=~0;
    store >> savedcrc;
    return (savedcrc==crc);
    }
  return false;
  }


// Load transparency
FXbool PNGDecoder::load_transparency(FXStream& store,FXuint nbytes){
  FXuint  savedcrc=0;
  FXuint  crc=~0;
  FXuchar a;

  crc=CRC32::CRC(crc,tRNS);

  switch(imagetype){
  case Gray:
    store.load(transparency_16bit,2);
    crc=CRC32::CRC(crc,transparency_16bit,2);
    switch(bitdepth){
    case  1:
      transparency=RGBGRAY(pal1bit[transparency_16bit[1]&0x1]);
      break;
    case  2:
      transparency=RGBGRAY(pal2bit[transparency_16bit[1]&0x3]);
      break;
    case  4:
      transparency=RGBGRAY(pal4bit[transparency_16bit[1]&0xf]);
      break;
    case  8:
    case 16:
      transparency=RGBGRAY(transparency_16bit[1]);
      break;
    default:
      return false;
      }
    break;
  case Indexed:
    if(colormap==nullptr || nbytes>ncolormap) return false;
    for(FXuint c=0; c<nbytes; c++){
      store >> a;
      crc=CRC32::CRC(crc,a);
      colormap[c] = (colormap[c]&0xffffff) | (a<<24);
      }
    break;
  case RGB:
    store.load(transparency_16bit,6);
    crc=CRC32::CRC(crc,transparency_16bit,6);
    transparency=FXRGB(transparency_16bit[1],transparency_16bit[3],transparency_16bit[5]);
    break;
  default:
    return false;
    }
  crc^=~0;
  store >> savedcrc;
  return (crc==savedcrc);
  }


FXbool PNGDecoder::load_data(FXStream& store,FXuint length){
  FXuint crc=~0;

  crc=CRC32::CRC(crc,IDAT);

  while(length || stream.avail_in){

    // fill up data
    if(BUFFERSIZE-stream.avail_in && length){
      FXuint n=FXMIN(BUFFERSIZE-stream.avail_in,length);
      store.load(buffer+stream.avail_in,n);
      crc=CRC32::CRC(crc,buffer+stream.avail_in,n);
      stream.avail_in += n;
      length-=n;
      if(length==0){
        FXuint storedcrc;
        store >> storedcrc;
        crc^=~0;
        if(storedcrc!=crc){
          FXTRACE((100,"png: data crc mismatch\n"));
          return false;
          }
        }
      }

    // Setup in and output buffers
    stream.next_in = buffer;
    if(stream.avail_out == 0){
      stream.next_out  = crow;
      stream.avail_out = (interlace == Adam7) ? irowbytes[imagepass]+1 : nrowbytes+1;
      }

    // decompress
    int zstatus = inflate(&stream,Z_NO_FLUSH);
    if(zstatus!=Z_OK){
      if(zstatus==Z_STREAM_END){
        FXTRACE((100,"png: end of stream\n"));
        inflateEnd(&stream);
        }
      else if(zstatus==Z_BUF_ERROR){
        FXTRACE((100,"png: need more input bytes\n"));
        return true;
        }
      else{
        inflateEnd(&stream);
        FXTRACE((100,"png: zlib error %d: %s\n",zstatus,stream.msg));
        return false;
        }
      }

    // completed row
    if(stream.avail_out==0){
      if(!process_row()){
        FXTRACE((100,"png: process row failed\n"));
        }
      }

    // update input buffer
    moveElms(buffer,stream.next_in,stream.avail_in);
    }
  return true;
  }


static inline FXuchar predictor(FXint a,FXint b,FXint c){
  const FXint p = a + b - c;
  const FXint pa = FXABS(p - a);
  const FXint pb = FXABS(p - b);
  const FXint pc = FXABS(p - c);
  if (pa <= pb && pa <= pc)
    return a;
  else if (pb <= pc)
    return b;
  else
    return c;
  }


void PNGDecoder::apply_filter(FXuchar type,FXuchar* current,const FXuchar* previous,FXint nbytes){
  if(previous){
    switch(type){
    case 0:
      break;
    case 1:
      for(FXint i=stride; i<nbytes; i++){
        current[i]+=current[i-stride];
        }
      break;
    case 2:
      for(FXint i=0; i<nbytes; i++){
        current[i]+=previous[i];
        }
      break;
    case 3:
      for(FXint i=0; i<stride; i++){
        current[i]+=((previous[i])/2);
        }
      for(FXint i=stride; i<nbytes; i++){
        current[i]+=((current[i-stride]+previous[i])/2);
        }
      break;
    case 4:
      for(FXint i=0; i<stride; i++){
        current[i]+=previous[i];
        }
      for(FXint i=stride; i<nbytes; i++){
        current[i]+=predictor(current[i-stride],previous[i],previous[i-stride]);
        }
      break;
    default:
      break;
      }
    }
  else{
    switch(type){
    case 0:
      break;
    case 1:
      for(FXint i=stride; i<nbytes; i++){
        current[i]+=current[i-stride];
        }
      break;
    case 2:
      break;
    case 3:
      for(FXint i=stride; i<nbytes; i++){
        current[i]+=((current[i-stride]) / 2);
        }
      break;
    case 4:
      for(FXint i=stride; i<nbytes; i++){
        current[i]+=current[i-stride];
        }
      break;
    default:
      break;
      }
    }
  }


FXbool PNGDecoder::process_row(){
  const FXuchar filtertype=crow[0];
  const FXuchar * pdata = imagerow ? prow+1 : nullptr;
  FXuchar* data = crow+1;

  if(filtertype>4){
    FXTRACE((100,"png: invalid filter type\n"));
    return false;
    }

  if(interlace==Adam7 && imagepass>=7){
    FXTRACE((100,"png: imagepass >= 7"));
    return false;
    }

  FXint nbytes = (interlace==Adam7) ? irowbytes[imagepass] : nrowbytes;

  apply_filter(filtertype,data,pdata,nbytes);

  if(interlace==Adam7){

    FXColor* dst=image+xoffset[imagepass]+((yoffset[imagepass]+(ystep[imagepass]*imagerow))*width);

    const FXColor* end=dst+width-xoffset[imagepass];

    copy_pixels(dst,end,data,nbytes,xstep[imagepass]);

    imagerow++;

    while(imagepass<7 && imagerow>=iheight[imagepass]){
      imagepass++;
      imagerow=0;
      }
    }
  else{
    if(imagerow>=height){
      FXTRACE((100,"png: imagerow >= height"));
      return false;
      }
    FXColor* dst=image+(imagerow*width);
    copy_pixels(dst,dst+width,data,nrowbytes,1);
    imagerow++;
    }


  // swap row pointers
  FXuchar* trow=crow;
  crow=prow;
  prow=trow;
  return true;
  }


void PNGDecoder::copy_pixels(FXColor* dst,const FXColor* end,const FXuchar* src,FXint nbytes,FXint step/*=1*/){
  switch(imagetype){
  case Gray:
    switch(bitdepth){
    case 1:
      for(FXint b=0; b<nbytes-1; b++){
        *dst=RGBGRAY(pal1bit[0x1&src[b]>>7]); dst+=step;
        *dst=RGBGRAY(pal1bit[0x1&src[b]>>6]); dst+=step;
        *dst=RGBGRAY(pal1bit[0x1&src[b]>>5]); dst+=step;
        *dst=RGBGRAY(pal1bit[0x1&src[b]>>4]); dst+=step;
        *dst=RGBGRAY(pal1bit[0x1&src[b]>>3]); dst+=step;
        *dst=RGBGRAY(pal1bit[0x1&src[b]>>2]); dst+=step;
        *dst=RGBGRAY(pal1bit[0x1&src[b]>>1]); dst+=step;
        *dst=RGBGRAY(pal1bit[0x1&src[b]]); dst+=step;
        }
      for(FXint s=7; s>=0 && dst<end; s--){
        *dst=RGBGRAY(pal1bit[0x1&src[nbytes-1]>>s]);
        dst+=step;
        }
      break;
    case 2:
      for(FXint b=0; b<nbytes-1; b++){
        *dst=RGBGRAY(pal2bit[0x3&src[b]>>6]); dst+=step;
        *dst=RGBGRAY(pal2bit[0x3&src[b]>>4]); dst+=step;
        *dst=RGBGRAY(pal2bit[0x3&src[b]>>2]); dst+=step;
        *dst=RGBGRAY(pal2bit[0x3&src[b]]); dst+=step;
        }
      for(FXint s=6; s>=0 && dst<end; s-=2){
        *dst=RGBGRAY(pal2bit[0x3&src[nbytes-1]>>s]);
        dst+=step;
        }
      break;
    case 4:
      for(FXint b=0; b<nbytes-1; b++){
        *dst=RGBGRAY(pal4bit[0xf&src[b]>>4]); dst+=step;
        *dst=RGBGRAY(pal4bit[0xf&src[b]]); dst+=step;
        }
      for(FXint s=4; s>=0 && dst<end; s-=4){
        *dst=RGBGRAY(pal4bit[0xf&src[nbytes-1]>>s]);
        dst+=step;
        }
      break;
    case 8:
      for(FXint b=0; b<nbytes; b++){
        *dst=FXRGB(src[b],src[b],src[b]); dst+=step;
        }
      break;
    case 16:
      if(transparency){
        for(FXint b=0; b<nbytes; b+=2){
          *dst=FXRGB(src[b],src[b],src[b]);
          if(memcmp(src+b,transparency_16bit,2)==0)
            *dst&=0xffffff;
          dst+=step;
          }
        }
      else{
        for(FXint b=0; b<nbytes; b+=2){
          *dst=FXRGB(src[b],src[b],src[b]);
          dst+=step;
          }
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    break;
  case GrayAlpha:
    switch(bitdepth){
    case 8:
      for(FXint b=0; b<nbytes; b+=2){
        *dst=FXRGBA(src[b],src[b],src[b],src[b+1]); dst+=step;
        }
      break;
    case 16:
      for(FXint b=0; b<nbytes; b+=4){
        *dst=FXRGBA(src[b],src[b],src[b],src[b+2]); dst+=step;
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    break;
  case Indexed:
    switch(bitdepth){
    case 1:
      for(FXint b=0; b<nbytes-1; b++){
        *dst=colormap[0x1&src[b]>>7]; dst+=step;
        *dst=colormap[0x1&src[b]>>6]; dst+=step;
        *dst=colormap[0x1&src[b]>>5]; dst+=step;
        *dst=colormap[0x1&src[b]>>4]; dst+=step;
        *dst=colormap[0x1&src[b]>>3]; dst+=step;
        *dst=colormap[0x1&src[b]>>2]; dst+=step;
        *dst=colormap[0x1&src[b]>>1]; dst+=step;
        *dst=colormap[0x1&src[b]]; dst+=step;
        }
      for(FXint s=7; s>=0 && dst<end; s--){
        *dst=colormap[0x1&src[nbytes-1]>>s]; dst+=step;
        }
      break;
    case 2:
      for(FXint b=0; b<nbytes-1; b++){
        *dst=colormap[0x3&src[b]>>6]; dst+=step;
        *dst=colormap[0x3&src[b]>>4]; dst+=step;
        *dst=colormap[0x3&src[b]>>2]; dst+=step;
        *dst=colormap[0x3&src[b]]; dst+=step;
        }
      for(FXint s=6; s>=0 && dst<end; s-=2){
        *dst=colormap[0x3&src[nbytes-1]>>s]; dst+=step;
        }
      break;
    case 4:
      for(FXint b=0; b<nbytes-1; b++){
        *dst=colormap[0xf&src[b]>>4]; dst+=step;
        *dst=colormap[0xf&src[b]]; dst+=step;
        }
      *dst=colormap[0xf&src[nbytes-1]>>4]; dst+=step;
      if(dst<end){
        *dst=colormap[0xf&src[nbytes-1]]; dst+=step;
        }
      break;
    case 8:
      for(FXint b=0; b<nbytes; b++){
        *dst=colormap[src[b]]; dst+=step;
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    break;
  case RGB:
    switch(bitdepth){
    case 8:
      for(FXint b=0; b<nbytes; b+=3){
        *dst=FXRGB(src[b+0],src[b+1],src[b+2]); dst+=step;
        }
      break;
    case 16:
      if(transparency){
        for(FXint b=0; b<nbytes; b+=6){
          *dst=FXRGB(src[b+0],src[b+2],src[b+4]);
          if(memcmp(src+b,transparency_16bit,6)==0)
            *dst&=0xffffff;
          dst+=step;
          }
        }
      else{
        for(FXint b=0; b<nbytes; b+=6){
          *dst=FXRGB(src[b+0],src[b+2],src[b+4]);
          dst+=step;
          }
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    break;
  case RGBA:
    switch(bitdepth){
    case 8:
      for(FXint b=0; b<nbytes; b+=4){
        *dst=FXRGBA(src[b+0],src[b+1],src[b+2],src[b+3]); dst+=step;
        }
      break;
    case 16:
      for(FXint b=0; b<nbytes; b+=8){
        *dst=FXRGBA(src[b+0],src[b+2],src[b+4],src[b+6]); dst+=step;
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    break;
  default:
    FXASSERT(0);
    break;
    }
  }


void PNGDecoder::apply_transparency(){
  if(transparency && (imagetype==RGB || imagetype==Gray) && bitdepth<16){
    for(FXuint i=0;i<width*height;i++){
      if(image[i]==transparency)
        image[i]&=0xffffff;
      }
    }
  }


// Prepare
FXbool PNGDecoder::prepare(){

  // allocate output image
  if(!allocElms(image,width*height)) return false;

  if(!prepare_stride()) return false;

  // allocate raw data buffer
  if(!allocElms(buffer,BUFFERSIZE)) return false;

  // allocate current scanline
  if(!allocElms(crow,1+nrowbytes)) return false;

  // allocate previous scanline
  if(!allocElms(prow,1+nrowbytes)) return false;

  if(interlace==Adam7){
    if(imagetype==Gray || imagetype==Indexed){
      for(FXint p=0; p<7; p++){
        if(xoffset[p]<width && yoffset[p]<height){
          irowbytes[p] = (7+((xstep[p]-1+width-xoffset[p])>>xshift[p])*bitdepth)>>3;
          iheight[p]   = (ystep[p]-1+height-yoffset[p])>>yshift[p];
          }
        else{
          irowbytes[p]=iheight[p]=0;
          }
        }
      }
    else{
      FXint ncomponents;
      switch(imagetype){
        case GrayAlpha: ncomponents=2; break;
        case RGB      : ncomponents=3; break;
        case RGBA     : ncomponents=4; break;
        default       : return false; break;
        }
      for(FXint p=0; p<7; p++){
        if(xoffset[p]<width && yoffset[p]<height){
          irowbytes[p] = (((xstep[p]-1+width-xoffset[p])>>xshift[p])*bitdepth*ncomponents)>>3;
          iheight[p]   = (ystep[p]-1+height-yoffset[p])>>yshift[p];
          }
        else {
          irowbytes[p]=iheight[p]=0;
          }
        }
      }

    while(irowbytes[imagepass]==0 && imagepass<7)
      imagepass++;

    if(imagepass==7){
      return false;
      }
    }

  // Init zlib decompressor
  inflateInit(&stream);
  return true;
  }


// Decode start
FXbool PNGDecoder::parse(FXStream& store,FXColor*& output_image,FXint& output_width,FXint& output_height){
  FXuint chunklength;
  FXuint chunkid;
  FXuint crc=~0;

  store.setBigEndian(true);

  // Load header
  if(!load_header(store)){
    FXTRACE((100, "png: invalid header\n"));
    return false;
    }

  // Check validity
  if(!is_valid()){
    FXTRACE((100, "png: invalid or unsupported\n"));
    return false;
    }


  // Allocate
  if(!prepare()){
    FXTRACE((100, "png: failed to prepare\n"));
    return false;
    }

  while(!store.eof()){

    // Get chunk
    store >> chunklength;
    store >> chunkid;

    // Palette
    if(chunkid==PLTE){
      if(!load_colormap(store,chunklength)){
        FXTRACE((100, "png: failed to load colormap\n"));
        return false;
        }
      }

    // Transparancy
    else if(chunkid==tRNS){
      if(!load_transparency(store,chunklength)){
        FXTRACE((100, "png: failed to load transparency\n"));
        return false;
        }
      }

    // Data
    else if(chunkid==IDAT){
      if(!load_data(store,chunklength)){
        FXTRACE((100, "png: failed to load data\n"));
        return false;
        }
      }
    else if(chunkid==IEND){
      crc^=~0;
      store >> crc;
      if((interlace==Adam7 && imagepass==7) || (interlace==NoInterlace && imagerow==height)){
        apply_transparency();
        output_image = image;
        output_width = width;
        output_height = height;
        image = nullptr;
        return true;
        }
      FXTRACE((100, "png: incomplete image (imagerow: %d imagepass: %d height: %d)\n",imagerow,imagepass,height));
      return false;
      }
    else{
      if(chunklength) store.position(chunklength,FXFromCurrent);
      crc^=~0;
      store >> crc;
      }
    }
  FXTRACE((100, "png: eof no IEND?\n"));
  return false;
  }


/*******************************************************************************/

// PNG Encoder
class PNGEncoder : public PNGBase {
protected:
  struct ColorHash {
    FXColor  color;
    FXushort index;
    ColorHash():index(0xffff){ }
    };
protected:
  const FXColor *image;
protected:
  FXuchar       *output;
  ColorHash     *colorhash;
protected:
  FXbool has_alpha() const;
  FXbool is_gray() const;
  FXbool is_indexed();
  FXuchar get_color_index(FXColor);
protected:
  void   save_header(FXStream &);
  void   save_iend(FXStream&);
  void   save_colormap(FXStream&);
  void   save_transparency(FXStream&,FXuint);
  FXbool save_data(FXStream&);
protected:
  void apply_filter(FXuchar type,FXuchar* out,const FXuchar* current,const FXuchar* previous);
  void copy_pixels(FXint row);
  void process_row(FXint row);
protected:
  FXbool unparse(FXStream & store);
  FXbool prepare();
public:
  PNGEncoder(const FXColor*,FXint w,FXint h);
  static FXbool save(FXStream & store,const FXColor *,FXint,FXint);
 ~PNGEncoder();
  };


// Construct encoder
PNGEncoder::PNGEncoder(const FXColor* img,FXint w,FXint h):PNGBase(w,h),image(img),output(nullptr),colorhash(nullptr){
  };


// Destroy encoder
PNGEncoder::~PNGEncoder(){
  freeElms(output);
  delete [] colorhash;
  };


FXbool PNGEncoder::save(FXStream& store,const FXColor* clr,FXint w,FXint h){
  PNGEncoder encoder(clr,w,h);
  return encoder.unparse(store);
  }


// Check for alpha
FXbool PNGEncoder::has_alpha() const {
  for(FXuint i=0;i<width*height;i++){
    if(FXALPHAVAL(image[i])<255) return true;
    }
  return false;
  }


// Check if greyscale
FXbool PNGEncoder::is_gray() const {
  for(FXuint i=0; i<width*height; i++){
    if(FXREDVAL(image[i])!=FXGREENVAL(image[i]) || FXGREENVAL(image[i])!=FXBLUEVAL(image[i])) return false;
    }
  return false;
  }


// Compare colors
static int compare_colors(const void *p,const void *q){
  FXColor x=*(const FXColor*)p;
  FXColor y=*(const FXColor*)q;
  if(x < y) return -1;
  if(x > y) return 1;
  return 0;
  }


// Get index for given color
FXuchar PNGEncoder::get_color_index(FXColor color){
  FXint p=color%337;
  FXint x=color%331+1;
  while(colorhash[p].index!=0xffff){
    if(colorhash[p].color==color){
      return colorhash[p].index;
      }
    p=(p+x)%337;
    }
  return 0;
  }


// Check color indexing
FXbool PNGEncoder::is_indexed(){
  ColorHash* indexhash=new ColorHash[337];
  FXint ncolors=0;
  FXint p,x;
  FXColor color;

  allocElms(colormap,256);

  for(FXuint i=0; i<width*height; i++){
    color=image[i];

    // Find position in table
    p=color%337;
    x=color%331+1;
    while(indexhash[p].index!=0xffff){         // Empty slot
      if(indexhash[p].color==color) goto nxt;
      p=(p+x)%337;
      }

    // If no more room in colormap, we failed
    if(ncolors>=256) {
      delete [] indexhash;
      freeElms(colormap);
      return false;
      }

    // Add new color
    indexhash[p].color=color;
    indexhash[p].index=ncolors;
    colormap[ncolors]=color;
    ncolors++;
nxt:continue;
    }

  // sort colors and update hash table
  ncolormap=ncolors;
  qsort(colormap,ncolormap,sizeof(FXColor),&compare_colors);

  // Remap indexes
  for(FXint i=0; i<ncolormap; i++){
    color=colormap[i];
    //fxmessage("%.3d: (%.3d,%.3d,%.3d,%.3d)\n",i,FXALPHAVAL(color),FXREDVAL(color),FXBLUEVAL(color),FXGREENVAL(color));
    p=color%337;
    x=color%331+1;
    while(indexhash[p].index!=0xffff){
      if(indexhash[p].color==colormap[i]){
        indexhash[p].index=i;
        goto nx2;
        }
      p=(p+x)%337;
      }
    FXASSERT(0);
nx2:continue;
    }
  colorhash=indexhash;
  return true;
  }


#if 0

FXbool PNGEncoder::is_indexed() const {
  FXint   ncolors=0;
  FXColor color;
  FXint   p,x;
  FXColor  colortable[337];             // Colors encountered in image
  FXushort mapindex[337];               // Map index assigned to color

  // Clear map index
  memset(mapindex,0xff,sizeof(mapindex));

  // Hash all colors from image
  for(FXint i=0; i<width*height; i++){

    // Get pixel
    color=image[i];

    // Find position in table
    p=color%337;
    x=color%331+1;
    while(mapindex[p]!=0xffff){         // Empty slot
      if(colortable[p]==color) goto nxt;
      p=(p+x)%337;
      }

    // If no more room in colormap, we failed
    if(ncolors>=256) return false;

    // Add new color
    colortable[p]=color;                // Add color to color hash table
    mapindex[p]=ncolors;                // Remember map index of this color
    ncolors++;
nxt:continue;
    }
  return true;
  }
#endif



// Filter
void PNGEncoder::apply_filter(FXuchar type,FXuchar* dest,const FXuchar* current,const FXuchar* previous){
  const FXint nbytes=nrowbytes;
  if(__likely(previous)){
    switch(type){
    case 0:
      for(FXint i=0; i<nbytes; i++){
        dest[i]=current[i];
        }
      break;
    case 1:
      for(FXint i=0; i<stride; i++){
        dest[i]=current[i];
        }
      for(FXint i=stride; i<nbytes; i++){
        dest[i]=current[i]-current[i-stride];
        }
      break;
    case 2:
      for(FXint i=0; i<nbytes; i++){
        dest[i]=current[i]-previous[i];
        }
      break;
    case 3:
      for(FXint i=0; i<stride; i++){
        dest[i]=current[i]-((previous[i])/2);
        }
      for(FXint i=stride; i<nbytes; i++){
        dest[i]=current[i]-((current[i-stride]+previous[i])/2);
        }
      break;
    case 4:
      for(FXint i=0; i<stride; i++){
        dest[i]=current[i]-previous[i];
        }
      for(FXint i=stride; i<nbytes; i++){
        dest[i]=current[i]-predictor(current[i-stride],previous[i],previous[i-stride]);
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    }
  else{
    switch(type){
    case 0:
      for(FXint i=0; i<nbytes; i++){
        dest[i]=current[i];
        }
      break;
    case 1:
      for(FXint i=0; i<stride; i++){
        dest[i]=current[i];
        }
      for(FXint i=stride; i<nbytes; i++){
        dest[i]=current[i]-current[i-stride];
        }
      break;
    case 2:
      for(FXint i=0; i<nbytes; i++){
        dest[i]=current[i];
        }
      break;
    case 3:
      for(FXint i=0; i<stride; i++){
        dest[i]=current[i];
        }
      for(FXint i=stride; i<nbytes; i++){
        dest[i]=current[i]-((current[i-stride]) / 2);
        }
      break;
    case 4:
      for(FXint i=0; i<stride; i++){
        dest[i]=current[i];
        }
      for(FXint i=stride; i<nbytes; i++){
        dest[i]=current[i]-current[i-stride];
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    }
  }


// Save colormap
void PNGEncoder::save_colormap(FXStream& store){
  FXuint chunklength=ncolormap*3;
  FXuint nalpha=0;
  FXuint crc=~0;

  // Calculate CRC
  crc=CRC32::CRC(crc,PLTE);

  store << chunklength;
  store << PLTE;
  for(FXint i=0; i<ncolormap; i++){
    const FXuchar r = FXREDVAL(colormap[i]);
    const FXuchar g = FXGREENVAL(colormap[i]);
    const FXuchar b = FXBLUEVAL(colormap[i]);
    const FXuchar a = FXALPHAVAL(colormap[i]);
    if(a<255) nalpha++;
    crc=CRC32::CRC(crc,r);
    crc=CRC32::CRC(crc,g);
    crc=CRC32::CRC(crc,b);
    store << r;
    store << g;
    store << b;
    }
  crc^=~0;
  store << crc;
  if(nalpha) save_transparency(store,nalpha);
  }


// Save transparancy
void PNGEncoder::save_transparency(FXStream& store,FXuint chunklength){
  FXuint crc=~0;

  // Calculate CRC
  crc=CRC32::CRC(crc,tRNS);

  store << chunklength;
  store << tRNS;
  for(FXuint i=0; i<chunklength; i++){
    const FXuchar a=FXALPHAVAL(colormap[i]);
    crc=CRC32::CRC(crc,a);
    store << a;
    }
  crc^=~0;
  store << crc;
  }


// Save IEND
void PNGEncoder::save_iend(FXStream& store){
  FXuint chunklength=0;
  FXuint crc=~0;

  // Calculate CRC
  crc=CRC32::CRC(crc,IEND);
  crc^=~0;
  store << chunklength;
  store << IEND;
  store << crc;
  }


// Save header
void PNGEncoder::save_header(FXStream& store){
  FXuint chunklength=13;
  FXuint crc=~0;

  // Start with PNG signature
  store.save(signature,ARRAYNUMBER(signature));

  // Calculate CRC
  crc=CRC32::CRC(crc,IHDR);
  crc=CRC32::CRC(crc,width);
  crc=CRC32::CRC(crc,height);
  crc=CRC32::CRC(crc,bitdepth);
  crc=CRC32::CRC(crc,imagetype);
  crc=CRC32::CRC(crc,compression);
  crc=CRC32::CRC(crc,filter);
  crc=CRC32::CRC(crc,interlace);
  crc^=~0;

  // Header Chunk
  store << chunklength;
  store << IHDR;
  store << width;
  store << height;
  store << bitdepth;
  store << imagetype;
  store << compression;
  store << filter;
  store << interlace;
  store << crc;
  }


// Copy pixels
void PNGEncoder::copy_pixels(FXint j){
  FXuchar * dst=crow;
  FXuint i;
  switch(imagetype){
  case Gray:
    for(i=0; i<width; i++){
      *dst++ = FXREDVAL(image[(j*width)+i]);
      }
    break;
  case GrayAlpha:
    for(i=0; i<width; i++){
      *dst++ = FXREDVAL(image[(j*width)+i]);
      *dst++ = FXALPHAVAL(image[(j*width)+i]);
      }
    break;
  case Indexed:
    switch(bitdepth){
/*
    case 2:
      for(i=0; i<width-3; i+=4){
        *dst++ = get_color_index(image[(j*width)+i+0])<<6 |
                 get_color_index(image[(j*width)+i+1])<<4 |
                 get_color_index(image[(j*width)+i+2])<<2;
                 get_color_index(image[(j*width)+i+3]);
        }
      if(i<width-3){ *dst |= get_color_index(image[(j*width)+i+0])<<6; i++; }
      if(i<width-2){ *dst |= get_color_index(image[(j*width)+i+1])<<4; i++; }
      if(i<width-1){ *dst |= get_color_index(image[(j*width)+i+2])<<2; i++; }
      if(i<width-0){ *dst |= get_color_index(image[(j*width)+i+3]); }
      break;
*/
    case 4:
      for(i=0; i<width-1; i+=2){
        *dst++ = get_color_index(image[(j*width)+i])<<4 | get_color_index(image[(j*width)+i+1]);
        }
      if(i<width) *dst++ = get_color_index(image[(j*width)+i])<<4;
      break;
    case 8:
      for(i=0; i<width; i++){
        *dst++ = get_color_index(image[(j*width)+i]);
        }
      break;
    default:
      FXASSERT(0);
      break;
      }
    break;
  case RGB:
    for(i=0; i<width; i++){
      *dst++ = FXREDVAL(image[(j*width)+i]);
      *dst++ = FXGREENVAL(image[(j*width)+i]);
      *dst++ = FXBLUEVAL(image[(j*width)+i]);
      }
    break;
  case RGBA:
    for(i=0; i<width; i++){
      *dst++ = FXREDVAL(image[(j*width)+i]);
      *dst++ = FXGREENVAL(image[(j*width)+i]);
      *dst++ = FXBLUEVAL(image[(j*width)+i]);
      *dst++ = FXALPHAVAL(image[(j*width)+i]);
      }
    break;
  default:
    FXASSERT(0);
    break;
    }
  }


// Process row
void PNGEncoder::process_row(FXint row){
  FXint sum;
  FXint min = (width*127);
  FXuchar desiredtype = 0;

  copy_pixels(row);

  for(FXuchar filtertype=0; filtertype<5; filtertype++){
    if(row)
      apply_filter(filtertype,output+1,crow,prow);
    else
      apply_filter(filtertype,output+1,crow,nullptr);

    sum = 0;
    for(FXuint b=0; b<nrowbytes; b++){
      sum+=FXABS((FXschar)output[1+b]);
      }

    if(sum<min){
      min=sum;
      desiredtype=filtertype;
      }
    }

  if(desiredtype<4){
    if(row)
      apply_filter(desiredtype,output+1,crow,prow);
    else
      apply_filter(desiredtype,output+1,crow,nullptr);
    }

  output[0]=desiredtype;
  }


FXbool PNGEncoder::save_data(FXStream & store){
  FXuval startpos=store.position();
  FXuint chunklength=0;
  FXuint crc=~0;

  store << chunklength;
  store << IDAT;

  // Init CRC
  crc=CRC32::CRC(crc,IDAT);

  for(FXuint j=0; j<height; j++){

    // process row
    process_row(j);

    // compress row
    stream.next_in=output;
    stream.avail_in=nrowbytes+1;
    while((stream.avail_in>0) || (j==height-1)){
      if(stream.avail_out==0){
        stream.next_out=buffer;
        stream.avail_out=BUFFERSIZE;
        }
      int zstatus=deflate(&stream,(j==height-1) ? Z_FINISH : Z_NO_FLUSH);
      if(zstatus==Z_OK){
        if(stream.avail_out==0){
          crc=CRC32::CRC(crc,buffer,BUFFERSIZE);
          store.save(buffer,BUFFERSIZE);
          }
        }
      else if(zstatus==Z_STREAM_END){
        if(BUFFERSIZE-stream.avail_out>0){
          crc=CRC32::CRC(crc,buffer,BUFFERSIZE-stream.avail_out);
          store.save(buffer,BUFFERSIZE-stream.avail_out);
          }
        deflateEnd(&stream);
        break;
        }
      else{
        deflateEnd(&stream);
        return false;
        }
      }

    // swap row pointers
    FXuchar* trow = crow;
    crow = prow;
    prow = trow;
    }

  // Start of Chunk
  crc^=~0;
  store << crc;
  FXuval endpos=store.position();
  chunklength=endpos-startpos-12;
  store.position(startpos,FXFromStart);
  store << chunklength;
  store.position(endpos,FXFromStart);
  return true;
  }


FXbool PNGEncoder::prepare(){

  // Make sure we have valid data
  if(image==nullptr || width<=0 || height<=0) return false;

  // Always 8 bits per pixel (for now)
  bitdepth=8;
  compression=Deflate;
  filter=Adaptive;
  interlace=NoInterlace;

  // Determine imagetype
  if(is_indexed()){
    imagetype=Indexed;
    if(ncolormap<=16) bitdepth=4;
    if(ncolormap<=4) bitdepth=2;
    //if (ncolormap <= 2)
    //  bitdepth = 1;
    }
  else{
    if(has_alpha()){
      if(is_gray())
        imagetype=GrayAlpha;
      else
        imagetype=RGBA;
      }
    else{
      if(is_gray())
        imagetype=Gray;
      else
        imagetype=RGB;
      }
    }

  prepare_stride();

  // allocate compressed data buffer
  if(!allocElms(buffer,BUFFERSIZE)) return false;

  // allocate current scanline
  if(!allocElms(crow,nrowbytes)) return false;

  // allocate previous scanline
  if(!allocElms(prow,nrowbytes)) return false;

  // allocate output
  if(!allocElms(output,1+nrowbytes)) return false;

  // Init zlib decompressor
  deflateInit(&stream,Z_DEFAULT_COMPRESSION);
  return true;
  }


FXbool PNGEncoder::unparse(FXStream& store){

  store.setBigEndian(true);

  if(!prepare()) return false;

  save_header(store);

  if(imagetype==Indexed){
    save_colormap(store);
    }

  if(!save_data(store)) return false;

  save_iend(store);
  return true;
  }


/*******************************************************************************/


// Check if stream contains a PNG
FXbool fxcheckPNG(FXStream& store){
  FXuchar sig[ARRAYNUMBER(signature)];
  store.load(sig,ARRAYNUMBER(signature));
  store.position(-ARRAYNUMBER(signature),FXFromCurrent);
  return equalElms(sig,signature,ARRAYNUMBER(signature));
  }


// Load a PNG image
FXbool fxloadPNG(FXStream& store,FXColor*& data,FXint& width,FXint& height){
  return PNGDecoder::load(store,data,width,height);
//  PNGDecoder decoder;
//  return decoder.parse(store,img,ww,hh);
  }


// Save a PNG image
FXbool fxsavePNG(FXStream& store,const FXColor* data,FXint width,FXint height){
  return PNGEncoder::save(store,data,width,height);
//  PNGEncoder encoder(data,width,height);
//  return encoder.unparse(store);
  }


/*******************************************************************************/


#else


// Check if stream contains a PNG
FXbool fxcheckPNG(FXStream&){
  return false;
  }


// Stub routine
FXbool fxloadPNG(FXStream&,FXColor*& data,FXint& width,FXint& height){
  static const FXColor color[2]={FXRGB(0,0,0),FXRGB(255,255,255)};
  static const FXuchar png_bits[] = {
   0xff, 0xff, 0xff, 0xff, 0x01, 0x00, 0x00, 0x80, 0xfd, 0xff, 0xff, 0xbf,
   0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0,
   0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0,
   0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0,
   0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0,
   0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0, 0xc5, 0x23, 0xc4, 0xa1,
   0x45, 0x24, 0x24, 0xa2, 0x45, 0x64, 0x24, 0xa0, 0x45, 0xa4, 0x24, 0xa0,
   0x45, 0x24, 0x25, 0xa0, 0xc5, 0x23, 0x26, 0xa3, 0x45, 0x20, 0x24, 0xa2,
   0x45, 0x20, 0x24, 0xa2, 0x45, 0x20, 0xc4, 0xa1, 0x05, 0x00, 0x00, 0xa0,
   0x05, 0x00, 0x00, 0xa0, 0x05, 0x00, 0x00, 0xa0, 0xfd, 0xff, 0xff, 0xbf,
   0x01, 0x00, 0x00, 0x80, 0xff, 0xff, 0xff, 0xff};
  allocElms(data,32*32);
  for(FXint p=0; p<32*32; p++){
    data[p]=color[(png_bits[p>>3]>>(p&7))&1];
    }
  width=32;
  height=32;
  return true;
  }


// Stub routine
FXbool fxsavePNG(FXStream&,const FXColor*,FXint,FXint){
  return false;
  }


#endif

}


